import hashlib
import os
import pickle

from coalib.misc import Constants


def get_data_path(log_printer, identifier):
    """
    Get the full path of ``identifier`` present in the user's data directory.

    :param log_printer: A LogPrinter object to use for logging.
    :param identifier:  The file whose path needs to be expanded.
    :return:            Full path of the file, assuming it's present in the
                        user's config directory.
                        Returns ``None`` if there is a ``PermissionError``
                        in creating the directory.
    """
    try:
        os.makedirs(Constants.USER_DATA_DIR, exist_ok=True)
        return os.path.join(Constants.USER_DATA_DIR, hash_id(identifier))
    except PermissionError:
        log_printer.err("Unable to create user data directory '{}'. Continuing"
                        ' without caching.'.format(Constants.USER_DATA_DIR))

    return None


def delete_files(log_printer, identifiers):
    """
    Delete the given identifiers from the user's coala data directory.

    :param log_printer: A LogPrinter object to use for logging.
    :param identifiers: The list of files to be deleted.
    :return:            True if all the given files were successfully deleted.
                        False otherwise.
    """
    error_files = []
    result = True
    for identifier in identifiers:
        try:
            file_path = get_data_path(log_printer, identifier)
            if os.path.isfile(file_path):
                os.remove(file_path)
            else:
                result = False
        except (OSError, TypeError) as e:
            error_files.append(hash_id(identifier))

    if len(error_files) > 0:
        error_files = ', '.join(error_files)
        log_printer.warn('There was a problem deleting the following '
                         'files: {}. Please delete them manually from '
                         "'{}'.".format(error_files, Constants.USER_DATA_DIR))
        result = False

    return result


def pickle_load(log_printer, identifier, fallback=None):
    """
    Get the data stored in ``filename`` present in the user
    config directory. Example usage:

    >>> from pyprint.NullPrinter import NullPrinter
    >>> from coalib.output.printers.LogPrinter import LogPrinter
    >>> log_printer = LogPrinter(NullPrinter())
    >>> test_data = {"answer": 42}
    >>> pickle_dump(log_printer, "test_project", test_data)
    True
    >>> pickle_load(log_printer, "test_project")
    {'answer': 42}
    >>> pickle_load(log_printer, "nonexistent_project")
    >>> pickle_load(log_printer, "nonexistent_project", fallback=42)
    42

    :param log_printer: A LogPrinter object to use for logging.
    :param identifier:  The name of the file present in the user config
                        directory.
    :param fallback:    Return value to fallback to in case the file doesn't
                        exist.
    :return:            Data that is present in the file, if the file exists.
                        Otherwise the ``default`` value is returned.
    """
    file_path = get_data_path(log_printer, identifier)
    if file_path is None or not os.path.isfile(file_path):
        return fallback
    with open(file_path, 'rb') as f:
        try:
            return pickle.load(f)
        except (pickle.UnpicklingError, EOFError) as e:
            log_printer.warn('The given file is corrupted and will be '
                             'removed.')
            delete_files(log_printer, [identifier])
            return fallback


def pickle_dump(log_printer, identifier, data):
    """
    Write ``data`` into the file ``filename`` present in the user
    config directory.

    :param log_printer: A LogPrinter object to use for logging.
    :param identifier:  The name of the file present in the user config
                        directory.
    :param data:        Data to be serialized and written to the file using
                        pickle.
    :return:            True if the write was successful.
                        False if there was a permission error in writing.
    """
    file_path = get_data_path(log_printer, identifier)
    if file_path is None:
        # Exit silently since the error has been logged in ``get_data_path``
        return False
    with open(file_path, 'wb') as f:
        pickle.dump(data, f)
    return True


def hash_id(text):
    """
    Hashes the given text.

    :param text: String to to be hashed
    :return:     A MD5 hash of the given string
    """
    return hashlib.md5(text.encode('utf-8')).hexdigest()


def get_settings_hash(sections,
                      targets=[],
                      ignore_settings: list=['disable_caching']):
    """
    Compute and return a unique hash for the settings.

    :param sections:        A dict containing the settings for each section.
    :param targets:         The list of sections that are enabled.
    :param ignore_settings: Setting keys to remove from sections before
                            hashing.
    :return:                A MD5 hash that is unique to the settings used.
    """
    settings = []
    for section in sections:
        if section in targets or targets == []:
            section_copy = sections[section].copy()
            for setting in ignore_settings:
                try:
                    section_copy.__getitem__(setting, ignore_defaults=True)
                    section_copy.delete_setting(setting)
                except IndexError:
                    continue
            settings.append(str(section_copy))

    return hash_id(str(settings))


def settings_changed(log_printer, settings_hash):
    """
    Determine if the settings have changed since the last run with caching.

    :param log_printer:   A LogPrinter object to use for logging.
    :param settings_hash: A MD5 hash that is unique to the settings used.
    :return:              Return True if the settings hash has changed
                          Return False otherwise.
    """
    project_hash = hash_id(os.getcwd())

    settings_hash_db = pickle_load(log_printer, 'settings_hash_db', {})
    if project_hash not in settings_hash_db:
        # This is the first time coala is run on this project, so the cache
        # will be flushed automatically.
        return False

    result = settings_hash_db[project_hash] != settings_hash
    if result:
        del settings_hash_db[project_hash]
        log_printer.debug('Since the configuration settings have '
                          'changed since the last run, the '
                          'cache will be flushed and rebuilt.')

    return result


def update_settings_db(log_printer, settings_hash):
    """
    Update the config file last modification date.

    :param log_printer:   A LogPrinter object to use for logging.
    :param settings_hash: A MD5 hash that is unique to the settings used.
    """
    project_hash = hash_id(os.getcwd())

    settings_hash_db = pickle_load(log_printer, 'settings_hash_db', {})
    settings_hash_db[project_hash] = settings_hash
    pickle_dump(log_printer, 'settings_hash_db', settings_hash_db)



from collections import Iterable, OrderedDict


def inverse_dicts(*dicts):
    """
    Inverts the dicts, e.g. {1: 2, 3: 4} and {2: 3, 4: 4} will be inverted
    {2: [1], 3: [2], 4: [3, 4]}. This also handles dictionaries
    with Iterable items as values e.g. {1: [1, 2, 3], 2: [3, 4, 5]} and
    {2: [1], 3: [2], 4: [3, 4]} will be inverted to
    {1: [1, 2], 2: [1, 3], 3: [1, 2, 4], 4: [2, 4], 5: [2]}.
    No order is preserved.

    :param dicts: The dictionaries to invert.
    :return:      The inversed dictionary which merges all dictionaries into
                  one.
    """
    inverse = {}

    for dictionary in dicts:
        for key, value in dictionary.items():
            if isinstance(value, Iterable):
                for item in value:
                    add_pair_to_dict(item, key, inverse)
            else:
                add_pair_to_dict(value, key, inverse)

    return inverse


def add_pair_to_dict(key, value, dictionary):
    """
    Add (key, value) pair to the dictionary. The value is added to a list of
    values for the key.
    """
    if key in dictionary:
        dictionary[key].append(value)
    else:
        dictionary[key] = [value]


def update_ordered_dict_key(dictionary, old_key, new_key):
    return OrderedDict(((new_key if k == old_key else k), v)
                       for k, v in dictionary.items())



def enum(*sequential, **named):
    enums = dict(zip(sequential, range(len(sequential))), **named)
    str_dict = enums.copy()
    enums['reverse'] = dict((value, key) for key, value in enums.items())
    enums['str_dict'] = str_dict

    return type('Enum', (), enums)



from coalib.bearlib.aspects import Root, Taste


@Root.subaspect
class Metadata:
    """
    This describes any aspect that is related to metadata that is not
    inside your source code.
    """


@Metadata.subaspect
class CommitMessage:
    """
    Your commit message is important documentation associated with your
    source code. It can help you to identify bugs (e.g. through
    `git bisect`) or find missing information about unknown source code
    (through `git blame`).

    Commit messages are also sometimes used to generate - or write
    manually - release notes.
    """


@CommitMessage.subaspect
class Emptiness:
    """
    Your commit message serves as important documentation for your source
    code.
    """
    class docs:
        example = '(no text at all)'
        example_language = 'English'
        importance_reason = """
        An empty commit message shows the lack of documentation for your
        change.
        """
        fix_suggestions = 'Write a commit message.'


@CommitMessage.subaspect
class Shortlog:
    """
    Your commit shortlog is the first line of your commit message. It is
    the most crucial part and summarizes the change in the shortest possible
    manner.
    """


@Shortlog.subaspect
class ColonExistence:
    """
    Some projects force to use colons in the commit message shortlog
    (first line).
    """
    class docs:
        example = """
        FIX: Describe change further
        context: Describe change further
        """
        example_language = 'English'
        importance_reason = """
        The colon can be a useful separator for a context (e.g. a filename) so
        the commit message makes more sense to the reader or a classification
        (e.g. FIX, ...) or others. Some projects prefer not using colons
        specifically: consistency is key.
        """
        fix_suggestions = """
        Add or remove the colon according to the commit message guidelines.
        """

    shortlog_colon = Taste[bool](
        'Whether or not the shortlog has to contain a colon.',
        (True, False), default=True)


@Shortlog.subaspect
class TrailingPeriod:
    """
    Some projects force not to use trailing periods in the commit
    message shortlog (first line).
    """
    class docs:
        example = """
        Describe change.
        Describe change
        """
        example_language = 'English'
        importance_reason = """
        Consistency is key to make messages more readable. Removing a trailing
        period can also make the message shorter by a character.
        """
        fix_suggestions = """
        Add or remove the trailing period according to the commit message
        guidelines.
        """

    shortlog_period = Taste[bool](
        'Whether or not the shortlog has to contain a trailing period.',
        (True, False), default=False)


@Shortlog.subaspect
class Tense:
    """
    Most projects have a convention on which tense to use in the commit
    shortlog (the first line of the commit message).
    """
    class docs:
        example = """
        Add file
        Adding file
        Added file
        """
        example_language = 'English'
        importance_reason = """
        Consistency is key to make messages more readable.
        """
        fix_suggestions = """
        Rephrase the shortlog into the right tense.
        """

    shortlog_tense = Taste[str](
        'The tense of the shortlog.',
        ('imperative', 'present continuous', 'past'),
        default='imperative')


@Shortlog.subaspect
class Length:
    """
    The length of your commit message shortlog (first line).
    """
    class docs:
        example = """
        Some people just write very long commit messages. Too long. "
        Even full sentences. And more of them, too!
        """
        example_language = 'English'
        importance_reason = """
        A good commit message should be quick to read and concise. Also, git
        and platforms like GitHub do cut away everything beyond 72, sometimes
        even 50 characters making any longer message unreadable.
        """
        fix_suggestions = """
        Try to compress your message:

        - Using imperative tense usually saves a character or two
        - Omitting a trailing period saves another character
        - Leave out unneeded words or details
        - Use common abbreviations like w/, w/o or &.
        """

    max_shortlog_length = Taste[int](
        'The maximal number of characters the shortlog may contain.',
        (50, 72, 80), default=72)


@Shortlog.subaspect
class FirstCharacter:
    """
    The first character of your commit message shortlog (first line) usually
    should be upper or lower case consistently.

    If the commit message contains a colon, only the first character after
    the colon will be checked.
    """
    class docs:
        example = """
        Add coverage pragma
        Compatability: Add coverage pragma
        add coverage pragma
        Compatability: add coverage pragma
        """
        example_language = 'English'
        importance_reason = """
        Consistent commit messages are easier to read through.
        """
        fix_suggestions = """
        Convert your first character to upper/lower case. If your message starts
        with an identifier, consider rephrasing. Usually starting with a verb is
        a good idea.
        """

    shortlog_starts_upper_case = Taste[bool](
        'Whether or not the shortlog (first line) of a commit message should '
        'start with an upper case letter consistently.',
        (True, False), default=True)


@CommitMessage.subaspect
class Body:
    """
    Your commit body may contain an elaborate description of your commit.
    """


@Body.subaspect
class Existence:
    """
    Forces the commit message body to exist (nonempty).
    """
    class docs:
        example = """
        aspects: Add CommitMessage.Body
        """
        example_language = 'English'
        importance_reason = """
        Having a nonempty commit body is important if you consistently want
        elaborate documentation on all commits.
        """
        fix_suggestions = """
        Write a commit message with a body.
        """


@Body.subaspect
class Length:
    """
    The length of your commit message body lines.
    """
    class docs:
        example = """
        Some people just write very long commit messages. Too long.
        Way too much actually. If they would just break their lines!
        """
        example_language = 'English'
        importance_reason = """
        Git and platforms like GitHub usually break everything beyond 72
        characters, making a message containing longer lines hard to read.
        """
        fix_suggestions = """
        Simply break your lines right before you hit the border.
        """

    max_body_length = Taste[int](
        'The maximal number of characters the body may contain in one line. '
        'The newline character at each line end does not count to that length.',
        (50, 72, 80), default=72)



from coala_utils.decorators import enforce_signature

from coalib.bearlib.languages import Languages


class TasteError(AttributeError):
    """
    A taste is not allowed to be accessed.
    """


class TasteMeta(type):
    """
    Metaclass for :class:`coalib.bearlib.aspectclasses.Taste`

    Allows defining taste cast type via :meth:`.__getitem__`, like:

    >>> Taste[int]().cast_type
    <class 'int'>
    """

    def __getitem__(cls, _cast_type):
        class Taste(cls):
            cast_type = _cast_type

        Taste.__name__ = Taste.__qualname__ = '%s[%s]' % (
            cls.__name__, _cast_type.__name__)
        return Taste


class Taste(metaclass=TasteMeta):
    """
    Defines tastes in aspectclass definitions.

    Tastes can be made only available for certain languages by
    providing a ``tuple`` of language identifiers on instantiation:

    >>> Taste[bool](
    ...     'Ignore ``using`` directives in C#.',
    ...     (True, False), default=False,
    ...     languages=('CSharp', )
    ... ).languages
    (C#,)

    If no `languages` are given, they will be available for any language.
    See :class:`coalib.bearlib.aspectclasses.Root` for further usage.
    """
    cast_type = str

    @enforce_signature
    def __init__(self, description: str='', suggested_values: tuple=(),
                 default=None, languages: tuple=()):
        """
        Creates a new taste that can be optionally only available for the
        given `languages`, which must be language identifiers supported by
        :class:`coalib.bearlib.languages.Language`.

        No need to specify name an cast type:

        The taste name is defined by the taste's attribute name in an
        aspectclass definition.

        The value cast type is defined via indexing on class level.
        """
        self.description = description
        self.suggested_values = suggested_values
        self.default = default
        self.languages = Languages(languages)

    def __get__(self, obj, owner=None):
        """
        Checks availability of taste for aspectclass instance `obj`'s
        ``.language`` before returning the specific taste value.
        """
        if obj is not None:
            # ==> access from aspectclass instance
            if self.languages and obj.language not in self.languages:
                raise TasteError('%s.%s is not available for %s.' % (
                    type(obj).__qualname__, self.name, obj.language))
            return obj.__dict__[self.name]
        # ==> access from aspectclass
        return self

    def __set__(self, obj, value):
        """
        Ensures that `value` is only set once in `obj`'s ``.__dict__``.
        """
        if self.name in obj.__dict__:
            raise AttributeError(
                "can't set taste values of aspectclass instances")
        obj.__dict__[self.name] = self.cast_type(value)
